"""Google Colab utilities."""
import os

from .env import set_osenv
from .logging import getLogger

logger = getLogger(__name__)


def mount_google_drive(
    workspace: str = None,
    project: str = None,
    mountpoint: str = "/content/drive",
    force_remount: bool = False,
    timeout_ms: int = 120000,
) -> None:
    """Mount Google Drive to Colab."""
    try:
        from google.colab import drive

        drive.mount(mountpoint, force_remount=force_remount, timeout_ms=timeout_ms)

        if isinstance(workspace, str):
            if not workspace.startswith(os.path.sep) and not workspace.startswith(".."):
                workspace = os.path.join(mountpoint, workspace)
            set_osenv("HYFI_WORKSPACE_ROOT", workspace)
            logger.info(f"Setting HYFI_WORKSPACE_ROOT to {workspace}")
        if isinstance(project, str):
            set_osenv("HYFI_PROJECT_NAME", project)
            logger.info(f"Setting HYFI_PROJECT_NAME to {project}")
    except ImportError:
        logger.warning("Google Colab not detected.")



import random
from pathlib import Path
from typing import Any, Optional, Union

from omegaconf import DictConfig
from pydantic import BaseModel, validator

from ..env import ProjectConfig, _to_config, _to_dict, getLogger
from ..hydra import _compose, _load, _merge, _methods, _print, _save, _save_json
from ..utils.lib import ensure_import_module

logger = getLogger(__name__)


class PathConfig(BaseModel):
    task_name: str = "default-task"
    root: str = None
    batch_name: str = None
    verbose: bool = False

    class Config:
        extra = "ignore"

    def __init__(self, **data: Any):
        if not data:
            data = _compose("path=__batch__")
            logger.info(
                "There are no arguments to initilize a config, using default config."
            )
        super().__init__(**data)

    @property
    def root_dir(self):
        # return as an absolute path
        return Path(self.root).absolute()

    @property
    def output_dir(self):
        return self.root_dir / "outputs"

    @property
    def batch_dir(self):
        return self.output_dir / self.batch_name

    @property
    def library_dir(self):
        return self.root_dir / "libs"

    @property
    def data_dir(self):
        return self.root_dir / "data"

    @property
    def model_dir(self):
        return self.root_dir / "models"

    @property
    def cache_dir(self):
        return self.root_dir / "cache"

    @property
    def tmp_dir(self):
        return self.root_dir / "tmp"

    @property
    def log_dir(self):
        log_dir = self.root_dir / "logs"
        log_dir.mkdir(parents=True, exist_ok=True)
        return log_dir


class BaseBatchConfig(BaseModel):
    batch_name: str
    batch_num: int = None
    output_dir: Path = Path.cwd() / "outputs"
    output_suffix: str = None
    output_extention: Optional[str] = ""
    random_seed: bool = True
    seed: int = None
    resume_run: bool = False
    resume_latest: bool = False
    num_workers: int = 1
    device: str = "cpu"
    num_devices: Optional[int] = None
    config_yaml = "config.yaml"
    config_json = "config.json"
    config_dirname = "configs"
    verbose: Union[bool, int] = False

    def __init__(self, **data):
        if not data:
            data = _compose("batch")
            logger.info(
                f"There is no batch in the config, using default batch: {data.batch_name}"
            )
        super().__init__(**data)
        self.init_batch_num()

    def init_batch_num(self):
        if self.batch_num is None:
            num_files = len(list(self.config_dir.glob(self.config_filepattern)))
            if self.resume_latest:
                self.batch_num = num_files - 1
            else:
                self.batch_num = num_files
        if self.verbose:
            logger.info(
                f"Init batch - Batch name: {self.batch_name}, Batch num: {self.batch_num}"
            )

    @validator("seed")
    def _validate_seed(cls, v, values):
        if values["random_seed"] or v is None or v < 0:
            random.seed()
            seed = random.randint(0, 2**32 - 1)
            if values.get("verbose"):
                logger.info(f"Setting seed to {seed}")
            return seed
        return v

    @validator("output_extention")
    def _validate_output_extention(cls, v):
        if v:
            return v.strip(".")
        else:
            return ""

    @property
    def batch_dir(self):
        batch_dir = self.output_dir / self.batch_name
        batch_dir.mkdir(parents=True, exist_ok=True)
        return batch_dir

    @property
    def config_dir(self):
        config_dir = self.batch_dir / self.config_dirname
        config_dir.mkdir(parents=True, exist_ok=True)
        return config_dir

    @property
    def file_prefix(self):
        return f"{self.batch_name}({self.batch_num})"

    @property
    def output_file(self):
        if self.output_suffix:
            return f"{self.file_prefix}_{self.output_suffix}.{self.output_extention}"
        else:
            return f"{self.file_prefix}.{self.output_extention}"

    @property
    def config_filename(self):
        return f"{self.file_prefix}_{self.config_yaml}"

    @property
    def config_jsonfile(self):
        return f"{self.file_prefix}_{self.config_json}"

    @property
    def config_filepattern(self):
        return f"{self.batch_name}(*)_{self.config_yaml}"

    @property
    def config_filepath(self):
        return self.config_dir / self.config_filename

    @property
    def config_jsonpath(self):
        return self.config_dir / self.config_jsonfile


class BaseConfigModel(BaseModel):
    config_name: str = None
    config_group: str = None
    name: str
    path: PathConfig = None
    project: ProjectConfig = None
    module: DictConfig = None
    auto: Union[DictConfig, str] = None
    force: Union[DictConfig, str] = None
    autoload: bool = False
    version: str = "0.0.0"
    _config_: DictConfig = None
    _initial_config_: DictConfig = None

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"
        validate_assignment = False
        exclude = {
            "_config_",
            "_initial_config_",
            "__data__",
            "path",
            "module",
            "secret",
            "auto",
            "project",
        }
        include = {}
        underscore_attrs_are_private = True
        property_set_methods = {
            "name": "set_name",
            "root_dir": "set_root_dir",
        }

    def __init__(
        self,
        config_name: str = None,
        config_group: str = None,
        root_dir: str = None,
        **args,
    ):
        if config_group is not None:
            args = _merge(_compose(config_group), args)
        else:
            args = _to_config(args)
        super().__init__(config_name=config_name, config_group=config_group, **args)

        object.__setattr__(self, "_config_", args)
        object.__setattr__(self, "_initial_config_", args.copy())
        self.initialize_configs(root_dir=root_dir)

    def __setattr__(self, key, val):
        super().__setattr__(key, val)
        method = self.__config__.property_set_methods.get(key)
        if method is not None:
            getattr(self, method)(val)

    def set_root_dir(self, root_dir: Union[str, Path]):
        path = self.config.path
        if path is None:
            path = _compose("path=_batch_")
            logger.info(
                f"There is no path in the config, using default path: {path.root}"
            )
            self._config_.path = path
        if root_dir is not None:
            path.root = str(root_dir)
        self.path = PathConfig(**path)

    def set_name(self, val):
        self._config_.name = val
        if self.name is None or self.name != val:
            self.name = val

    def initialize_configs(self, root_dir=None, **kwargs):
        self.root_dir = root_dir

    @property
    def config(self):
        return self._config_

    @property
    def root_dir(self) -> Path:
        return Path(self.path.root)

    @property
    def output_dir(self):
        return self.path.output_dir

    @property
    def project_name(self):
        return self.project.project_name

    @property
    def project_dir(self):
        return Path(self.project.project_dir)

    @property
    def workspace_dir(self):
        return Path(self.project.workspace_dir)

    @property
    def model_dir(self):
        return self.path.model_dir

    @property
    def log_dir(self):
        return self.project.path.log_dir

    @property
    def cache_dir(self):
        cache_dir = Path(self.project.path.cache)
        if cache_dir is None:
            cache_dir = self.output_dir / ".cache"
            cache_dir.mkdir(parents=True, exist_ok=True)
        return Path(cache_dir)

    @property
    def library_dir(self):
        return self.path.library_dir

    @property
    def verbose(self):
        return self.project.verbose

    def autorun(self):
        return _methods(self.auto, self)

    def show_config(self):
        _print(self.dict())

    def load_modules(self):
        """Load the modules"""
        if self.module.get("modules") is None:
            logger.info("No modules to load")
            return
        library_dir = self.library_dir
        for module in self.module.modules:
            name = module.name
            libname = module.libname
            liburi = module.liburi
            specname = module.specname
            libpath = library_dir / libname
            syspath = module.get("syspath")
            if syspath is not None:
                syspath = library_dir / syspath
            ensure_import_module(name, libpath, liburi, specname, syspath)

    def reset(self, objects=None):
        """Reset the memory cache"""
        if isinstance(objects, list):
            for obj in objects:
                del obj
        try:
            from ..utils.gpu import GPUMon

            GPUMon.release_gpu_memory()
        except ImportError:
            pass


class BaseBatchModel(BaseConfigModel):
    batch: BaseBatchConfig = None

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"
        validate_assignment = False
        exclude = {
            "_config_",
            "_initial_config_",
            "__data__",
            "path",
            "module",
            "secret",
            "auto",
            "project",
        }
        include = {}
        underscore_attrs_are_private = True
        property_set_methods = {
            "name": "set_name",
            "batch_name": "set_name",
            "batch_num": "set_batch_num",
            "root_dir": "set_root_dir",
        }

    def __init__(self, **args):
        super().__init__(**args)

    def set_name(self, val):
        super().set_name(val)
        self._config_.batch.batch_name = val
        self.batch.batch_name = val
        self.initialize_configs(name=val)

    def set_batch_num(self, val):
        self._config_.batch.batch_num = val
        self.batch.batch_num = val

    def initialize_configs(
        self, root_dir=None, batch_config_class=BaseBatchConfig, **kwargs
    ):
        super().initialize_configs(root_dir=root_dir, **kwargs)

        self.batch = batch_config_class(**self.config.batch)
        self.batch_num = self.batch.batch_num
        # if self.project.use_huggingface_hub:
        #     self.secrets.init_huggingface_hub()
        if self.verbose:
            logger.info(
                f"Initalized batch: {self.batch_name}({self.batch_num}) in {self.root_dir}"
            )

    @property
    def batch_name(self):
        return self.batch.batch_name

    @property
    def batch_num(self):
        return self.batch.batch_num

    @property
    def seed(self):
        return self.batch.seed

    @property
    def batch_dir(self):
        return self.batch.batch_dir

    @property
    def data_dir(self):
        return self.path.data_dir

    @property
    def verbose(self):
        return self.batch.verbose

    @property
    def device(self):
        return self.batch.device

    @property
    def num_devices(self):
        return self.batch.num_devices

    def save_config(
        self,
        config=None,
        exclude=None,
        include=None,
    ):
        """Save the batch config"""
        if config is not None:
            self._config_ = config
        logger.info(f"Saving config to {self.batch.config_filepath}")
        cfg = _to_dict(self.config)
        if exclude is None:
            exclude = self.__config__.exclude

        if include:
            args = {}
            if isinstance(include, str):
                include = [include]
            for key in include:
                args[key] = cfg[key]
        else:
            args = cfg
            if exclude:
                if isinstance(exclude, str):
                    exclude = [exclude]
                for key in exclude:
                    args.pop(key, None)
        _save(args, self.batch.config_filepath)
        self.save_settings(exclude=exclude)
        return self.batch.config_filename

    def save_settings(self, exclude=None, exclude_none=True):
        def dumper(obj):
            if isinstance(obj, DictConfig):
                return _to_dict(obj)
            return str(obj)

        if exclude is None:
            exclude = self.__config__.exclude
        config = self.dict(exclude=exclude, exclude_none=exclude_none)
        if self.verbose:
            logger.info(f"Saving config to {self.batch.config_jsonpath}")
        _save_json(config, self.batch.config_jsonpath, default=dumper)

    def load_config(
        self,
        batch_name=None,
        batch_num=None,
        **args,
    ):
        """Load the config from the batch config file"""
        if self.verbose:
            logger.info(
                f"> Loading config for batch_name: {batch_name} batch_num: {batch_num}"
            )
        # self.config.batch.batch_num = batch_num
        if batch_name is None:
            batch_name = self.batch_name

        if batch_num is not None:
            cfg = self._initial_config_.copy()
            self.batch.batch_name = batch_name
            self.batch.batch_num = batch_num
            _path = self.batch.config_filepath
            if _path.is_file():
                logger.info(f"Loading config from {_path}")
                batch_cfg = _load(_path)
                if self.verbose:
                    logger.info("Merging config with the loaded config")
                cfg = _merge(cfg, batch_cfg)
            else:
                logger.info(f"No config file found at {_path}")
                batch_num = None
        else:
            cfg = self.config

        if self.verbose:
            logger.info(f"Merging config with args: {args}")
        self._config_ = _merge(cfg, args)

        self.batch_num = batch_num
        self.batch_name = batch_name

        return self.config

    def show_config(self, batch_name=None, batch_num=None):
        self.load_config(batch_name, batch_num)
        _print(self.dict())



import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import hydra
from omegaconf import DictConfig, ListConfig, OmegaConf, SCMode
from pydantic import BaseModel, BaseSettings, SecretStr, root_validator, validator
from pydantic.env_settings import SettingsSourceCallable

from .utils.batch import batcher
from .utils.env import load_dotenv
from .utils.logging import getLogger, setLogger
from .utils.notebook import is_notebook, load_extentions, set_matplotlib_formats

logger = getLogger(__name__)

__hydra_version_base__ = "1.2"
__config_module__ = "hyfi.conf"


def __version__():
    """Returns the version of HyFI"""
    from ._version import __version__

    return __version__


def _select(
    cfg: Any,
    key: str,
    *,
    default: Any = None,
    throw_on_resolution_failure: bool = True,
    throw_on_missing: bool = False,
):
    key = key.replace("/", ".")
    return OmegaConf.select(
        cfg,
        key=key,
        default=default,
        throw_on_resolution_failure=throw_on_resolution_failure,
        throw_on_missing=throw_on_missing,
    )


def _to_dict(
    cfg: Any,
):
    if isinstance(cfg, dict):
        cfg = _to_config(cfg)
    if isinstance(cfg, (DictConfig, ListConfig)):
        return OmegaConf.to_container(
            cfg,
            resolve=True,
            throw_on_missing=False,
            structured_config_mode=SCMode.DICT,
        )
    return cfg


def _to_config(
    cfg: Any,
):
    return OmegaConf.create(cfg)


class AboutConfig(BaseModel):
    """About Configuration"""

    name: str = "HyFI"
    author: str = "Young Joon Lee"
    description: str = (
        "Hydra Fast Interface (Hydra and Pydantic based interface framework)"
    )
    website: str = "https://hyfi.entelecheia.cc"
    version: str = __version__()


__about__ = AboutConfig()


class DistFramwork(BaseModel):
    """Distributed Framework Configuration"""

    backend: str = "joblib"
    initialize: bool = False
    num_workers: int = 1


class BatcherConfig(BaseModel):
    """Batcher Configuration"""

    procs: int = 1
    minibatch_size: int = 1_000
    backend: str = "joblib"
    task_num_cpus: int = 1
    task_num_gpus: int = 0
    verbose: int = 10


class JobLibConfig(BaseModel):
    """JobLib Configuration"""

    config_name: str = "__init__"
    distributed_framework: DistFramwork = DistFramwork()
    batcher: BatcherConfig = BatcherConfig()
    __initilized__: bool = False

    class Config:
        extra = "allow"
        underscore_attrs_are_private = True

    def __init__(
        self,
        config_name: str = "__init__",
        **data: Any,
    ):
        if not data:
            data = _compose(f"joblib={config_name}", config_module=__config_module__)
            logger.debug(
                "There are no arguments to initilize a config, using default config."
            )
        super().__init__(config_name=config_name, **data)

    def init_backend(
        self,
    ):
        """Initialize the backend for joblib"""
        backend = self.distributed_framework.backend

        if self.distributed_framework.initialize:
            backend_handle = None
            if backend == "ray":
                import ray

                ray_cfg = {"num_cpus": self.distributed_framework.num_workers}
                logger.info(f"initializing ray with {ray_cfg}")
                ray.init(**ray_cfg)
                backend_handle = ray

            elif backend == "dask":
                from dask.distributed import Client

                dask_cfg = {"n_workers": self.distributed_framework.num_workers}
                logger.info(f"initializing dask client with {dask_cfg}")
                client = Client(**dask_cfg)
                logger.debug(client)

            batcher.batcher_instance = batcher.Batcher(
                backend_handle=backend_handle, **self.batcher.dict()
            )
            logger.info(f"initialized batcher with {batcher.batcher_instance}")
        self.__initilized__ = True

    def stop_backend(self):
        """Stop the backend for joblib"""
        backend = self.distributed_framework.backend
        if batcher.batcher_instance:
            logger.info("stopping batcher")
            del batcher.batcher_instance

        logger.info("stopping distributed framework")
        if self.distributed_framework.initialize:
            if backend == "ray":
                import ray

                if ray.is_initialized():
                    ray.shutdown()
                    logger.info("shutting down ray")

            # elif modin_engine == 'dask':
            #     from dask.distributed import Client

            #     if Client.initialized():
            #         client.close()
            #         log.info(f'shutting down dask client')


class PathConfig(BaseModel):
    config_name: str = "__init__"
    dotenv_path: str = None
    workspace: str = None
    project: str = "hyfi-default"
    data: str = None
    home: str = None
    hyfi: str = None
    resources: str = None
    runtime: str = None
    archive: str = None
    corpus: str = None
    datasets: str = None
    logs: str = None
    models: str = None
    outputs: str = None
    cache: str = None
    tmp: str = None
    library: str = None
    verbose: bool = False

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True

    def __init__(
        self,
        config_name: str = "__init__",
        **data: Any,
    ):
        if not data:
            data = _compose(f"path={config_name}", config_module=__config_module__)
            logger.debug(
                "There are no arguments to initilize a config, using default config."
            )
        super().__init__(config_name=config_name, **data)

    @property
    def log_dir(self):
        Path(self.logs).mkdir(parents=True, exist_ok=True)
        return Path(self.logs).absolute()

    @property
    def cache_dir(self):
        Path(self.cache).mkdir(parents=True, exist_ok=True)
        return Path(self.cache).absolute()


class DotEnvConfig(BaseSettings):
    """Environment variables for HyFI"""

    # Internal
    HYFI_WORKSPACE_ROOT: Optional[str]
    HYFI_PROJECT_NAME: Optional[str]
    HYFI_TASK_NAME: Optional[str]
    HYFI_PROJECT_ROOT: Optional[str]
    HYFI_DATA_ROOT: Optional[str]
    HYFI_RESOURCE_DIR: Optional[str]
    HYFI_LOG_LEVEL: Optional[str]
    HYFI_VERBOSE: Optional[Union[bool, str, int]]
    NUM_WORKERS: Optional[int]
    CACHED_PATH_CACHE_ROOT: Optional[str]
    # For other packages
    CUDA_DEVICE_ORDER: Optional[str]
    CUDA_VISIBLE_DEVICES: Optional[str]
    WANDB_PROJECT: Optional[str]
    WANDB_DISABLED: Optional[str]
    WANDB_DIR: Optional[str]
    WANDB_NOTEBOOK_NAME: Optional[str]
    WANDB_SILENT: Optional[Union[bool, str]]
    LABEL_STUDIO_SERVER: Optional[str]
    KMP_DUPLICATE_LIB_OK: Optional[str] = "True"
    TOKENIZERS_PARALLELISM: Optional[bool] = False
    # API Keys and Tokens
    WANDB_API_KEY: Optional[SecretStr]
    HUGGING_FACE_HUB_TOKEN: Optional[SecretStr]
    ECOS_API_KEY: Optional[SecretStr]
    FRED_API_KEY: Optional[SecretStr]
    NASDAQ_API_KEY: Optional[SecretStr]
    HF_USER_ACCESS_TOKEN: Optional[SecretStr]
    LABEL_STUDIO_USER_TOKEN: Optional[SecretStr]

    class Config:
        env_prefix = ""
        env_nested_delimiter = "__"
        case_sentive = False
        env_file = ".env"
        env_file_encoding = "utf-8"
        validate_assignment = True
        extra = "allow"

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            load_dotenv()
            return env_settings, file_secret_settings, init_settings

    @root_validator()
    def _check_and_set_values(cls, values):
        workspace = values.get("HYFI_WORKSPACE_ROOT")
        project = values.get("HYFI_PROJECT_NAME")
        if workspace is not None and project is not None:
            project_dir = os.path.join(workspace, "projects", project)
            values["HYFI_PROJECT_ROOT"] = project_dir
        for k, v in values.items():
            if v is not None:
                old_value = os.getenv(k.upper())
                if old_value is None or old_value != str(v):
                    os.environ[k.upper()] = str(v)
                    logger.debug(f"Set environment variable {k.upper()}={v}")
        return values


class ProjectConfig(BaseModel):
    """Project Config"""

    config_name: str = "__init__"
    project_name: str = "hyfi-project"
    task_name: str = None
    workspace_root: str = None
    project_root: str = None
    description: str = None
    use_huggingface_hub: bool = False
    use_wandb: bool = False
    verbose: bool = False
    # Config Classes
    dotenv: DotEnvConfig = None
    joblib: JobLibConfig = None
    path: PathConfig = None

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True

    def __init__(
        self,
        config_name: str = "__init__",
        **data: Any,
    ):
        if not data:
            data = _compose(f"project={config_name}", config_module=__config_module__)
            logger.debug(
                "There are no arguments to initilize a config, using default config."
            )
        super().__init__(config_name=config_name, **data)

    @validator("project_name", allow_reuse=True)
    def _validate_project_name(cls, v):
        if v is None:
            raise ValueError("Project name must be specified.")
        return v

    @property
    def workspace_dir(self):
        return Path(self.path.workspace)

    @property
    def project_dir(self):
        return Path(self.path.project)

    @property
    def version(self):
        return __about__.version

    def init_project(self):
        self.dotenv = DotEnvConfig()
        if self.path is None:
            self.path = PathConfig()
        if self.joblib is None:
            self.joblib = JobLibConfig()

        if self.dotenv.HYFI_VERBOSE is not None:
            self.verbose = self.dotenv.HYFI_VERBOSE
        self.dotenv.HYFI_DATA_ROOT = str(self.path.data)
        self.dotenv.CACHED_PATH_CACHE_ROOT = str(self.path.cache_dir / "cached_path")
        wandb_dir = str(self.path.log_dir)
        self.dotenv.WANDB_DIR = wandb_dir
        project_name = self.project_name.replace("/", "-").replace("\\", "-")
        self.dotenv.WANDB_PROJECT = project_name
        task_name = self.task_name.replace("/", "-").replace("\\", "-")
        notebook_name = self.path.log_dir / f"{task_name}-nb"
        notebook_name.mkdir(parents=True, exist_ok=True)
        self.dotenv.WANDB_NOTEBOOK_NAME = str(notebook_name)
        self.dotenv.WANDB_SILENT = str(not self.verbose)
        if self.use_wandb:
            try:
                import wandb

                wandb.init(project=self.project_name)
            except ImportError:
                logger.warning(
                    "wandb is not installed, please install it to use wandb."
                )
        if self.use_huggingface_hub:
            self.init_huggingface_hub()

    def init_huggingface_hub(self):
        """Initialize huggingface_hub"""
        try:
            from huggingface_hub import notebook_login
            from huggingface_hub.hf_api import HfFolder
        except ImportError:
            logger.warning(
                "huggingface_hub is not installed, please install it to use huggingface_hub."
            )
            return

        self.dotenv = DotEnvConfig()
        if (
            self.dotenv.HUGGING_FACE_HUB_TOKEN is None
            and self.dotenv.HF_USER_ACCESS_TOKEN is not None
        ):
            self.dotenv.HUGGING_FACE_HUB_TOKEN = self.dotenv.HF_USER_ACCESS_TOKEN

        local_token = HfFolder.get_token()
        if local_token is None:
            if is_notebook():
                notebook_login()
            else:
                logger.info(
                    "huggingface_hub.notebook_login() is only available in notebook,"
                    "set HUGGING_FACE_HUB_TOKEN manually"
                )


def _check_and_set_value(key, value):
    """Check and set value to environment variable"""
    env_key = key.upper()
    if value is not None:
        old_value = os.getenv(env_key, "")
        if str(old_value).lower() != str(value).lower():
            os.environ[env_key] = str(value)
            logger.debug("Set environment variable %s=%s", env_key, str(value))
    return value


class HyfiConfig(BaseModel):
    """HyFI config primary class"""

    hyfi_package_config_path: str = "pkg://hyfi.conf"
    hyfi_config_module: str = __config_module__
    hyfi_user_config_path: str = None

    debug_mode: bool = False
    print_config: bool = False
    print_resolved_config: bool = False
    verbose: bool = False
    logging_level: str = "WARNING"

    hydra: DictConfig = None

    about: AboutConfig = AboutConfig()
    project: ProjectConfig = None

    __version__: str = __version__()
    __initilized__: bool = False

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        validate_assignment = True
        extra = "allow"

    @root_validator()
    def _check_and_set_values(cls, values):
        key = "hyfi_package_config_path"
        val = _check_and_set_value(key, values.get(key))
        values[key] = val
        if val is not None:
            key = "hyfi_config_module"
            values[key] = _check_and_set_value(key, val.replace("pkg://", ""))
        return values

    @validator("hyfi_user_config_path")
    def _validate_hyfi_user_config_path(cls, v):
        return _check_and_set_value("hyfi_user_config_path", v)

    @validator("logging_level")
    def _validate_logging_level(cls, v, values):
        verbose = values.get("verbose", False)
        if verbose and v == "WARNING":
            v = "INFO"
        logger.setLevel(v)
        return v

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.about = __about__

    def init_workspace(
        self,
        workspace=None,
        project=None,
        task=None,
        log_level=None,
        autotime=True,
        retina=True,
        verbose=None,
        **kwargs,
    ):
        """Initialize project in notebook"""
        envs = DotEnvConfig(HYFI_VERBOSE=verbose)
        if isinstance(workspace, str):
            envs.HYFI_WORKSPACE_ROOT = workspace
        if isinstance(project, str):
            envs.HYFI_PROJECT_NAME = project
        if isinstance(task, str):
            envs.HYFI_TASK_NAME = task
        if isinstance(log_level, str):
            envs.HYFI_LOG_LEVEL = log_level
            setLogger(log_level)
            logger.setLevel(log_level)
        if autotime:
            load_extentions(exts=["autotime"])
        if retina:
            set_matplotlib_formats("retina")
        self.initialize()

    def initialize(self, config: Union[DictConfig, Dict] = None):
        """Initialize hyfi config"""
        if self.__initilized__:
            return
        if config is None:
            config = _compose(
                overrides=["+project=__init__"], config_module=__config_module__
            )
            logger.debug("Using default config.")

        if "project" not in config:
            logger.warning(
                "No project config found, skip project config initialization."
            )
            return
        self.project = ProjectConfig(**config["project"])
        self.project.init_project()
        self.project.joblib.init_backend()
        self.__initilized__ = True

    def terminate(self):
        """Terminate hyfi config"""
        if not self.__initilized__:
            return
        if self.project is not None:
            self.project.joblib.stop_backend()
        self.__initilized__ = False

    def __repr__(self):
        return f"HyFIConfig(project={self.project})"

    def __str__(self):
        return self.__repr__()

    @property
    def app_version(self):
        return self.about.app_version


__global_config__ = HyfiConfig()


class Dummy:
    def __call__(self, *args, **kwargs):
        return Dummy()


def _compose(
    config_group: str = None,
    overrides: List[str] = [],
    *,
    return_as_dict: bool = False,
    throw_on_resolution_failure: bool = True,
    throw_on_missing: bool = False,
    config_name: str = None,
    config_module: str = None,
    verbose: bool = False,
) -> Union[DictConfig, Dict]:
    """
    Compose your configuration from config groups and overrides (overrides=["override_name"])

    :param overrides: List of overrides to apply
    :param config_group: Config group name to select ('config_group=name')
    :param return_as_dict: Return the composed config as a dict
    :param throw_on_resolution_failure: Throw if resolution fails
    :param throw_on_missing: Throw if a config is missing
    :param config_name: Name of the config to compose
    :param verbose: Print the composed config

    :return: The composed config
    """
    config_module = config_module or __global_config__.hyfi_config_module
    # if verbose:
    logger.info("config_module: %s", config_module)
    is_initialized = hydra.core.global_hydra.GlobalHydra.instance().is_initialized()
    if config_group:
        _task = config_group.split("=")
        if len(_task) == 2:
            key, value = _task
        else:
            key = _task[0]
            value = "default"
        config_group = f"{key}={value}"
    else:
        key = None
        value = None
    if key and value:
        if is_initialized:
            cfg = hydra.compose(config_name=config_name, overrides=overrides)
        else:
            with hydra.initialize_config_module(
                config_module=config_module, version_base=__hydra_version_base__
            ):
                cfg = hydra.compose(config_name=config_name, overrides=overrides)
        cfg = _select(
            cfg,
            key=key,
            default=None,
            throw_on_missing=False,
            throw_on_resolution_failure=False,
        )
        if cfg is not None:
            overide = config_group
        else:
            overide = f"+{config_group}"
        if overrides:
            overrides.append(overide)
        else:
            overrides = [overide]
    # if verbose:
    logger.info(f"compose config with overrides: {overrides}")
    if is_initialized:
        if verbose:
            logger.info("Hydra is already initialized")
        cfg = hydra.compose(config_name=config_name, overrides=overrides)
    else:
        with hydra.initialize_config_module(
            config_module=config_module, version_base=__hydra_version_base__
        ):
            cfg = hydra.compose(config_name=config_name, overrides=overrides)

    if key and key != "task":
        cfg = _select(
            cfg,
            key=key,
            default=None,
            throw_on_missing=throw_on_missing,
            throw_on_resolution_failure=throw_on_resolution_failure,
        )
    logger.debug("Composed config: %s", OmegaConf.to_yaml(cfg))
    if return_as_dict and isinstance(cfg, DictConfig):
        return _to_dict(cfg)
    return cfg



"""Logging utilities"""
import logging
import os
import sys
import warnings


def setLogger(level=None, force=True, filterwarnings_action="ignore", **kwargs) -> None:
    """Set the logging level and format"""
    level = level or os.environ.get("HYFI_LOG_LEVEL", "INFO")
    level = level.upper()
    os.environ["HYFI_LOG_LEVEL"] = level
    if filterwarnings_action is not None:
        warnings.filterwarnings(filterwarnings_action)
    if isinstance(level, str):
        level = getattr(logging, level.upper(), logging.INFO)
    if sys.version_info >= (3, 8):
        logging.basicConfig(level=level, force=force, **kwargs)
    else:
        logging.basicConfig(level=level, **kwargs)


def getLogger(
    _name=None,
    _log_level=None,
    _fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
) -> logging.Logger:
    """Get a logger with a given name and log level"""
    _name = _name or __name__
    logger = logging.getLogger(_name)
    _log_level = _log_level or os.environ.get("HYFI_LOG_LEVEL", "INFO")
    logger.setLevel(_log_level)
    return logger



"""Environment variable utilities"""
import os
from pathlib import Path
from typing import Any

import dotenv
import hydra

from ..io.file import is_dir
from .logging import getLogger

logger = getLogger(__name__)


def getcwd():
    """Get the original working directory before hydra changed it"""
    try:
        return hydra.utils.get_original_cwd()
    except ValueError:
        return os.getcwd()


def dotenv_values(dotenv_path=None, **kwargs):
    """Load dotenv file and return a dict of key/value pairs"""
    config = dotenv.dotenv_values(dotenv_path=dotenv_path, **kwargs)
    return dict(config)


def load_dotenv(
    override: bool = False,
    dotenv_dir: str = None,
    dotenv_filename: str = ".env",
    verbose: bool = False,
) -> None:
    """Load dotenv file from the given directory or from the current directory"""
    dotenv_dir = dotenv_dir or getcwd()
    dotenv_path = Path(dotenv_dir, dotenv_filename)
    if dotenv_path.is_file():
        dotenv.load_dotenv(dotenv_path=dotenv_path, verbose=verbose, override=override)
        if verbose:
            logger.info("Loaded .env from %s", dotenv_path)
    else:
        if verbose:
            logger.info(
                "No .env file found in %s, finding .env in parent dirs", dotenv_path
            )
        dotenv_path = dotenv.find_dotenv()
        if dotenv_path:
            dotenv.load_dotenv(
                dotenv_path=dotenv_path, verbose=verbose, override=override
            )
            if verbose:
                logger.info("Loaded .env from %s", dotenv_path)
        else:
            if verbose:
                logger.info("No .env file found in %s", dotenv_path)


def get_osenv(key: str = None, default: str = None) -> Any:
    """Get the value of an environment variable or return the default value"""
    load_dotenv()
    if key:
        return os.environ.get(key, default)
    return os.environ


def set_osenv(key: str, value: Any) -> None:
    """Set the value of an environment variable"""
    if value and is_dir(value):
        value = os.path.abspath(value)
    pre_val = os.environ.get(key)
    if pre_val:
        logger.info("Overwriting %s=%s with %s", key, pre_val, value)
    else:
        logger.info("Setting %s=%s", key, value)
    os.environ[key] = value



"""HyFI: Hydra Fast Interface (Hydra and Pydantic based interface framework)"""
from .cli import getLogger, hydra_main
from .env import __about__ as about
from .env import __global_config__ as global_config
from .env import __hydra_version_base__
from .main import HyFI



"""Command line interface for HyFI"""
import os

import hydra

from .env import HyfiConfig, __hydra_version_base__
from .main import DictConfig, HyFI, getLogger

logger = getLogger(__name__)


def cmd(**args):
    """Run the command defined in the config file"""
    HyFI.run(args)


def about(**args):
    """Print the about information"""
    cfg = HyfiConfig(**args)
    name = cfg.about.name
    print()
    for k, v in cfg.about.dict().items():
        print(f"{k:11} : {v}")
    print(f"\nExecute `{name} --help` to see what you can do with {name}")


def cli_main(cfg: DictConfig) -> None:
    """Main function for the command line interface"""
    hyfi = HyfiConfig(**cfg)
    verbose = hyfi.verbose
    app_name = hyfi.about.name
    print_config = hyfi.print_config
    print_resolved_config = hyfi.print_resolved_config

    if verbose:
        print("## Command Line Interface for %s ##" % app_name)
    HyFI.initialize(cfg)

    if print_config:
        if print_resolved_config:
            logger.info("## hydra configuration resolved ##")
            HyFI.pprint(cfg)
        else:
            logger.info("## hydra configuration ##")
            print(HyFI.to_yaml(cfg))

    if verbose:
        logger.info(f"Hydra working directory : {os.getcwd()}")
        logger.info(f"Orig working directory  : {hydra.utils.get_original_cwd()}")

    HyFI.instantiate(cfg)

    HyFI.terminate()


def hydra_main() -> None:
    """Main function for the command line interface of Hydra"""
    hydra.main(
        config_path="conf", config_name="config", version_base=__hydra_version_base__
    )(cli_main)()


if __name__ == "__main__":
    """Run the command line interface"""
    hydra_main()
